<?php

namespace InSegment\ApiCore\Models;

use InSegment\ApiCore\Traits\Instantiatible;

use InSegment\ApiCore\Services\Transactor;
use InSegment\ApiCore\Services\BufferedBuilder;

abstract class SupplementaryTransactionTable extends SupplementaryTable
{
    use Instantiatible;
    
    const PUSH_MODE_INSERT = 0;
    const PUSH_MODE_UPDATE = 1;
    
    /**
     * Log push query
     *
     * @var array [
     *    string $class => \Illuminate\Database\Eloquent\Builder
     * ]
     */
    protected static $pushQueries = [];
    
    /**
     * Original table name
     *
     * @var string
     */
    protected $defaultTable;
    
    /**
     * UID
     *
     * @var string
     */
    public $uid;
    
    /**
     * Boot method
     * Subscribe to instantiation event
     * 
     * @return null
     */
    public static function boot()
    {
        static::instantiated(function ($model, $instantiator) {
            if (isset($instantiator['model']) && $instantiator['model']->uid !== $model->uid) {
                $model->uid = $instantiator['model']->uid;
                $model->updateTable();
            }
        });
    }
    
    /**
     * Constructor
     * 
     * @param array $attributes
     * @param string|null $uid
     */
    public function __construct(array $attributes = array(), $uid = null)
    {
        parent::__construct($attributes);
        $this->defaultTable = $this->getTable();
        $this->uid = $uid ?? Transactor::getTransactionUID();
        $this->updateTable();
    }
    
    /**
     * Update table of the model
     * 
     * @return $this
     */
    protected function updateTable()
    {
        $database = Transactor::getTransactionSchema();
        $table = $this->tableNameGenerator($this->uid);
        $this->table = "{$database}.{$table}";
        return $this;
    }

    /**
     * Table name generator
     * 
     * @return string
     */
    protected function tableNameGenerator()
    {
        return isset($this->uid) ? substr(implode('_', [$this->defaultTable, $this->uid]), 0, 64) : '';
    }
    
    /**
     * Create a new Eloquent query builder for the model.
     *
     * @param  \Illuminate\Database\Query\Builder  $query
     * @return \InSegment\ApiCore\Services\BufferedBuilder|static
     */
    public function newEloquentBuilder($query)
    {
        return new BufferedBuilder($query);
    }
    
    /**
     * Re-use the same query to push many records for the table
     * 
     * @param array $data
     * @param int $mode
     * @return null
     */
    public static function pushLog(array $data, int $mode)
    {
        if (!isset(static::$pushQueries[static::class])) {
            self::$pushQueries[static::class] = static::query();
        }
        
        switch($mode) {
            case self::PUSH_MODE_INSERT: self::$pushQueries[static::class]->insert($data); break;
            case self::PUSH_MODE_UPDATE: self::$pushQueries[static::class]->update($data); break;
            default: throw new \Exception('Invalid push mode specified for pushLog of SupplementaryTable');
        }
    }
    
    /**
     * Clear the push query
     * 
     * @return null
     */
    public static function clearPushQuery()
    {
        unset(static::$pushQueries[static::class]);
    }

}
